; ******************************************************
; **       Librera de rutinas para Amstrad CPC/GB    **
; **	   Ral Simarro, 	  Artaburu 2007       **
; ******************************************************
;
; Addapte pour le ZX81 par XavSnap.

; datos necesarios que vienen en la pila:
; ORIGEN	HL
; DESTINO	DE


;------- TASM ASM mnemonics. -------
; Compile this file using:
; Set TASMOPTS = -b
; tasm -80 ThisCode.tas MyBinary.BIN
;-----------------------------------
; Zx81 Program name: 

#define ORG  .org       ; TASM cross-assembler definitions
#define equ  .equ
#define db  .db
#define dw .word 
;-----------------------------------

#include "ZX81.sym" ; Load the symbol file.

ORG 16514

	pop af
	pop de
	pop hl

	; ****** Unpack pucrunch data ******

pucrunch:
	LD DE,(Dest_ADD) ; Entry Source packed data
	push   de        ; destination pointer to 2nd register  

        LD HL,(Sour_ADD) ; Destination for unpacked data

	exx             ; set  
	pop   de   
	push   hl
	push   de 
	push   bc 
	push   af 
	exx		; read the header self-modifying the
			; parameters straight into the code 
			; skip useless data 
	ld bc,6
	add hl,bc       ; skip "**PU**"
	ld   a, (hl) 	; starting escape 
	inc   hl  
	ld   (escPu+1), a 
	inc   hl        ; skip useless data "X"
	inc   hl  
	ld   a, (hl)       ; number of escape bits  
	inc   hl  
	ld   (escPub0+1), a 
	ld   (escPub1+1), a 
	ld   b, a          ; 8 - escape bits 
	ld   a, 8  
	sub   b 
	ld   (noescPu+1), a 
	ld   a, (hl)       ; maxGamma + 1 
	inc   hl  
	ld   (MaxGamma+1), a 
	ld   b, a          ; 8 - maxGamma  
	ld   a, 9
	sub   b  
	ld   (Max8Gamma+1), a 
	ld   a, (hl)       ; (1 << maxGamma) 
	inc   hl  
	ld   (Max1Gamma+1), a  
	add   a, a         ; (2 << maxGamma) - 1  
	dec   a  
	ld   (Max2Gamma+1), a   
	ld   a, (hl)       ; extra lz77_0 position bits
	inc   hl 
	ld   (ExtraBits+1), a 
	inc   hl           ; skip useless data 
	inc   hl
	ld   e, (hl) ; RLE table length  
	ld   (tablePu+1), hl       ; RLE table pointer 
	inc   hl
	ld   d, 0
	add   hl, de 
	ld   c, $80        ; start decompression 
	jr   loop_u 

newescPu:
	  ld   a, (escPu+1)  ; save old escape code  
	  ld   d, a 

escPub0:
	  ld   b, 2        ; ** parameter 
	  xor   a          ; get new escape code  
	  call   get_bits  
	  ld   (escPu+1), a  
	  ld   a, d 

noescPu:
	  ld   b, 6         ; ** parameter  
	  call   get_bits      ; get more bits to complete a byte 
	  exx            ; output the byte 
	  ld   (de), a 
	  inc   de 
	  exx 

loop_u:
	  xor   a 

escPub1:
	  ld   b, 2         ; ** parameter  
	  call   get_bits      ; get escape code 

escPu:
	  cp   0         ; ** parameter   
	  jr   nz, noescPu   
	  call   get_gamma      ; get length   
	  exx 
	  ld   b, 0   
	  ld   c, a   
	  exx   
	  cp   1   
	  jr   nz, lz77_0      ; lz77_0   
	  xor   a   
	  call   get_bit   
	  jr   nc, lz77_0_2    ; 2-byte lz77_0 
	  call   get_bit   
	  jr   nc, newescPu    ; escaped literal byte 
	  call   get_gamma     ; get length  
	  exx
	  ld   b, 1 
	  ld   c, a 
	  exx 

Max1Gamma:
	  cp   64              ; ** parameter   
	  jr   c, chrcode      ; short RLE, get bytecode 
Max8Gamma:
	  ld   b, 2            ; ** parameter
	  call   get_bits      ; complete length LSB
	  ex   af, af'  
	  call   get_gamma     ; length MSB  
	  exx  
	  ld   b, a   
	  ex   af, af'  
	  ld   c, a
	  exx 
chrcode:
	  call   get_gamma     ; get byte to repeat  
	  push   hl 
tablePu:
	  ld   hl, $0000       ; ** parameter   
	  ld   d, 0   
	  ld   e, a   
	  add   hl, de   
	  cp   32   
	  ld   a, (hl)  
	  pop   hl   
	  jr   c, dorle   
	  ld   a, e      ; get 3 more bits to complete the   
	  ld   b, 3      ; byte   
	  call   get_bits 
dorle:
	  exx            ; output the byte n times   
	  inc   c 
dorlei:
	  ld   (de), a   
	  inc   de   
	  dec   c   
	  jr   nz, dorlei   
	  dec   b   
	  jr   nz, dorlei   
	  exx
	  jr   loop_u 
lz77_0:
	  call   get_gamma ; offset MSB 
Max2Gamma:
	  cp   127         ; ** parameter   
	  ; ret   z   
	  jr   z, fin      ; EOF, return   
	  dec   a          ; (1...126 -> 0...125) 
ExtraBits:
	  ld   b, 0        ; ** parameter   
	  call   get_bits  ; complete offset MSB 
lz77_0_2:
	  ex   af, af'   
	  ld   b, 8        ; offset LSB   
	  call   get_bits   
	  cpl              ; xor'ed by the compressor   
	  exx              ; combine them into offset   
	  ld   l, a   
	  ex   af, af'   
	  ld   h, a   
	  inc   hl   
	  xor   a          ; CF = 0   
	  push   de        ; (current output position) - (offset) 
	  ex   de, hl  
	  sbc   hl, de
	  pop   de
	  inc   bc
	  ldir              ; copy
	  exx   
	  jp   loop_u 

;## Get a bit from the source stream. 
;## Return    CF = result 
get_bit:
	  sla   c         ; shift next bit into CF   
	  ret   nz   
	  ld   c, (hl)    ; get next byte   
	  inc   hl        ; increase source stream pointer   
	  rl   c          ; shift next bit into CF, bit0 = 1   
	  ret 

;## Get multiple bits from the source stream. 
;## In        B = number of bits to get 
;## Return    A = result 
get_bits:
	  dec   b   
	  ret   m   
	  sla   c         ; shift next bit into CF   
	  jr   nz, gb1   
	  ld   c, (hl)    ; get next byte   
	  inc   hl        ; increase source stream pointer   
	  rl   c          ; shift next bit into CF, bit0 = 1 
gb1:
          rla             ; rotate next bit into A 
	  jr   get_bits 

;## Get an Elias Gamma coded value from the source stream. 
;## Return    A = result 
get_gamma:
	  ld   b, 1 
MaxGamma:
	  ld   a, 7       ; ** parameter 
gg1:
	  call   get_bit  ; get bits until 0-bit or max 
	  jr   nc, gg2
	  inc   b  
	  cp   b
	  jr   nz, gg1 
gg2:
	  ld   a, 1         ; get the actual value
	  dec   b
	  jr   get_bits 

fin:   ; Restauramos los registros dobles y vuelta limpia
          exx
;	  LD HL,(Dest_ADD)
;	  EX DE,HL
;	  SBC HL,DE
;	  LD (PIC_LEN+1),HL
	  pop   af
	  pop   bc
	  pop   de
	  pop   hl
	  exx     ;ret   
	  ret   

Sour_ADD: ; Entry Source packed data
	dw Entry_DATAs

Dest_ADD: ; Destination for unpacked data
	dw $4400
LoadScr:
	LD HL,(Dest_ADD)
	LD DE,(RAM_D_FILE)
	INC DE
PIC_LEN:
	LD BC,$0318
	LDIR
	RET

Entry_DATAs:
.end